@@ -2,6 +2,7 @@
 """A filesystem-like interface for static and dynamic data.
 This can be used to e.g. browse REST APIs.
 """
+from contextlib import contextmanager
 from enum import Enum
 from pickle import dumps, loads
 from pprint import pformat
@@ -232,6 +233,9 @@ class FileSystem:
 
         self.post_cd_hook()
 
+    def cd_up(self):
+        self.cd(Option.up.value)
+
     @property
     def semantic_path(self) -> Path:
         """Convert indices in path to semantic values.
@@ -271,10 +275,10 @@ class FileSystem:
         return fs
 
     def __getitem__(self, k):
-        return self.root[k]
+        return self.get(k)
 
     def __contains__(self, k):
-        return k in self.root
+        return k in self.state.ls()
 
     ############################################################################
     # Internals
@@ -381,3 +385,21 @@ class FileSystem:
                     results = [result]
 
             yield from results
+
+
+@contextmanager
+def cd(filesystem: FileSystem, *keys: str):
+    """Change directory and finally reset the current directory.
+    Stay in the current directory by default. 
+    """
+    if not keys:
+        keys = (Option.stay.value)
+
+    path = filesystem.full_path
+    try:
+        filesystem.cd(*keys)
+        yield
+    except IndexError:
+        raise KeyError(keys)
+    finally:
+        filesystem.cd(*path)
