@@ -1,8 +1,9 @@
 from asyncio import CancelledError
 from cmd import Cmd
 from collections import defaultdict
+from contextlib import contextmanager
 from dataclasses import asdict
-from itertools import chain
+from itertools import repeat
 from json import dumps, loads
 from operator import contains
 from typing import Any, Callable, Dict, Iterable, List, Tuple, Union
@@ -11,30 +12,29 @@ import shlex
 import subprocess
 
 from mash import io_util
-from mash.filesystem.filesystem import FileSystem
+from mash.filesystem.filesystem import FileSystem, cd
 from mash.io_util import log, shell_ready_signal, print_shell_ready_signal, check_output
 from mash.shell import delimiters
-from mash.shell.delimiters import DEFINE_FUNCTION, LEFT_ASSIGNMENT, RIGHT_ASSIGNMENT
+from mash.shell.delimiters import DEFINE_FUNCTION, IF, LEFT_ASSIGNMENT, RETURN, RIGHT_ASSIGNMENT, THEN
+from mash.shell.env import ENV, Environment, show
+from mash.shell.errors import ShellError, ShellPipeError
 from mash.shell.function import InlineFunction
-from mash.util import for_any, has_method, identity, is_globbable, is_valid_method_name, match_words, omit_prefixes, split_prefixes, split_sequence, glob
+from mash.shell.parsing import expand_variables, expand_variables_inline, filter_comments, infer_infix_args, parse_commands
+from mash.util import for_any, has_method, identity, is_valid_method_name, omit_prefixes, removeprefix, split_prefixes, translate_terms
 
 
 confirmation_mode = False
 default_session_filename = '.shell_session.json'
 
+FALSE = ''
+TRUE = '1'
+COMMENT = '#'
+INNER_SCOPE = 'inner_scope'
 
 Command = Callable[[Cmd, str], str]
 Types = Union[str, bool, int, float]
 
 
-class ShellError(RuntimeError):
-    pass
-
-
-class ShellPipeError(RuntimeError):
-    pass
-
-
 class BaseShell(Cmd):
     """Extend CMD with various capabilities.
     This class is restricted to functionality that requires Cmd methods to be overrride.
@@ -73,10 +73,10 @@ class BaseShell(Cmd):
         # defaults
         self.ignore_invalid_syntax = True
 
-        self.env = {}
-        self.locals = FileSystem(defaultdict(dict))
+        self.locals = FileSystem(scope())
+        self.init_current_scope()
 
-        self.update_env(env)
+        self.env = Environment(self.locals)
 
         self.auto_save = False
         self.auto_reload = False
@@ -90,6 +90,10 @@ class BaseShell(Cmd):
         if self.auto_reload:
             self.try_load_session()
 
+    def init_current_scope(self):
+        self.locals.set(IF, [])
+        self.locals.set(ENV, {})
+
     @property
     def delimiters(self):
         """Return the most recent values of the delimiters.
@@ -118,16 +122,6 @@ class BaseShell(Cmd):
         self._do_char_method = method
         self._chars_allowed_for_char_method = chars
 
-    def update_env(self, env: Dict[str, Any] = None):
-        if env is None:
-            return
-
-        for k in self.env:
-            if k not in env:
-                env[k] = self.env[k]
-
-        self.env = env
-
     def eval(self, args: Iterable[str], quote=True) -> Types:
         """Evaluate / run `args` and return the result.
         """
@@ -135,7 +129,9 @@ class BaseShell(Cmd):
             args = (shlex.quote(arg) for arg in args)
             # args = ' '.join(shlex.quote(arg) for arg in args)
 
-        args = ' '.join(args)
+        args = ' '.join(filter_comments(args))
+        if not args:
+            return ''
 
         k = '_eval_output'
         line = f'{args} |> export {k}'
@@ -162,110 +158,6 @@ class BaseShell(Cmd):
 
         raise RuntimeError('Cannot retrieve result')
 
-    def handle_set_env_variable(self, lhs: Tuple[str], *rhs: str) -> str:
-        if len(lhs) != 1:
-            raise ShellError()
-
-        k = lhs[0]
-        self.set_env_variable(k, *rhs)
-        return ''
-
-    def handle_define_inline_function(self, terms: List[str]) -> str:
-        f, *args = terms
-        args = [arg for arg in args if arg != '']
-
-        if not args or not args[0].startswith('(') or not args[-1].endswith(')'):
-            lhs = ' '.join(terms)
-            raise ShellError(f'Invalid syntax for inline function: {lhs}')
-
-        # strip braces
-        if args[0] == '()':
-            args = []
-        else:
-            args[0] = args[0][1:]
-            args[-1] = args[-1][:-1]
-
-        self.locals.set(DEFINE_FUNCTION,
-                        InlineFunction('', *args, func_name=f))
-
-    def _save_inline_function(self) -> str:
-        func = self.locals[DEFINE_FUNCTION]
-        self.locals.rm(DEFINE_FUNCTION)
-
-        f = func.func_name
-        args = func.args
-        inner = func.command
-
-        if has_method(self, f'do_{f}'):
-            raise ShellError(
-                f'Name conflict: Cannot define inline function {f}, '
-                f'because there already exists a method do_{f}.')
-
-        if not is_valid_method_name(f):
-            raise ShellError(f'Invalid function name format: {f}')
-
-        inner = self.expand_variables_inline(inner)
-
-        # TODO use custom class with attr .functions instead of a string
-        self.locals['functions'][f] = InlineFunction(inner, *args)
-
-        positionals = ' '.join(args)
-        log(f'function {f}({positionals});')
-
-    def call_inline_function(self, f: InlineFunction, *args: str):
-        translations = {}
-
-        if len(args) != len(f.args):
-            raise ShellError(
-                f'Invalid number of arguments: {len(f.args)} arguments expected .')
-
-        for i, k in enumerate(f.args):
-            # quote item to preserve `\n`
-            translations[k] = shlex.quote(args[i])
-
-        terms = [term for term in f.command.split(' ') if term != '']
-        terms = list(self.translate_terms(terms, translations))
-
-        first_func = terms[0]
-        if not has_method(self, f'do_{first_func}') \
-                and first_func not in self.locals['functions']:
-            terms = ['print'] + terms
-
-        return self.eval(terms, quote=False)
-
-    def set_env_variable(self, k: str, *values: str):
-        """Set the variable `k` to `values`
-        """
-        if not is_valid_method_name(k):
-            raise ShellError(f'Invalid variable name format: {k}')
-
-        log(f'set {k}')
-        self.env[k] = ' '.join(values)
-        return k
-
-    def set_env_variables(self, keys: str, result: str):
-        """Set the variable `k` to `values`
-        """
-        if result is None:
-            raise ShellError(f'Missing return value in assignment: {keys}')
-
-        # TODO set multiple keys based on pattern matching
-        # e.g. i j = range(2)
-        k = keys[0]
-
-        self.env[k] = result
-
-    def show_env(self, env=None):
-        if env is None:
-            env = self.env
-
-        if not env:
-            return
-
-        print('Env')
-        for k in env:
-            print(f'\t{k}: {env[k]}')
-
     def onecmd_prehook(self, line):
         """Similar to cmd.precmd but executed before cmd.onecmd
         """
@@ -277,53 +169,23 @@ class BaseShell(Cmd):
 
         return line
 
-    def save_session(self, session=default_session_filename):
-        self.save_session_prehook()
+    def none(self, _: str) -> str:
+        """Do nothing. Similar to util.none.
+        This is a default value for self._do_char_method.
+        """
+        return ''
 
-        if not self.env:
-            logging.info('No env data to save')
-            return
+    def _save_result(self, value, overwrite=True):
+        if overwrite:
+            self._last_results = [value]
+        else:
+            self._last_results.append(value)
 
-        with open(session, 'w') as f:
-            try:
-                json = dumps(self.env)
-            except TypeError:
-                logging.debug('Cannot serialize self.env')
-                try:
-                    json = dumps(self.env, skip_keys=True)
-                except TypeError:
-                    json = dumps(asdict(self.env))
+    def is_function(self, func_name: str) -> bool:
+        return has_method(self, f'do_{func_name}') or self.is_inline_function(func_name)
 
-            f.write(json)
-
-    def try_load_session(self, session=default_session_filename):
-        self.load_session(session, strict=False)
-
-    def load_session(self, session: str = None, strict=True):
-        try:
-            with open(session) as f:
-                data = f.read()
-
-        except OSError as e:
-            if strict:
-                raise ShellError(e)
-
-            log(f'Session file not found: {session}: {e}')
-            return
-
-        if not data:
-            logging.info('No env data found')
-            return
-
-        env = loads(data)
-
-        log(f'Using session: {session}')
-        self.show_env(env)
-
-        # TODO handle key conflicts
-        self.update_env(env)
-
-        self.load_session_posthook()
+    def is_inline_function(self, func_name: str) -> bool:
+        return func_name in self.env and isinstance(self.env[func_name], InlineFunction)
 
     ############################################################################
     # Commands: do_*
@@ -340,7 +202,7 @@ class BaseShell(Cmd):
 
         for k in keys:
             if not is_valid_method_name(k):
-                raise ShellError('Invalid variable name format: {k}')
+                raise ShellError(f'Invalid variable name format: {k}')
 
         if LEFT_ASSIGNMENT in self.locals:
             assignee = self.locals[LEFT_ASSIGNMENT]
@@ -367,14 +229,21 @@ class BaseShell(Cmd):
         k, *values = args.split(' ')
 
         if len(values) == 0:
-            log(f'unset {k}')
+            values = ['']
+
+        self.set_env_variable(k, *values)
+
+    def do_unset(self, args: str):
+        """Unset keys
+        `unset [KEY [KEY..]]
+        """
+        for k in args.split(' '):
             if k in self.env:
                 del self.env[k]
             else:
                 logging.warning('Invalid key')
-            return
 
-        self.set_env_variable(k, *values)
+        return ''
 
     def do_shell(self, args):
         """System call
@@ -436,7 +305,7 @@ class BaseShell(Cmd):
 
     def do_map(self, args: str, delimiter='\n'):
         """Apply a function to every line.
-        If `$` is present, then each line from stdin is inserted there. 
+        If `$` is present, then each line from stdin is inserted there.
         Otherwise each line is appended.
 
         Usage
@@ -512,18 +381,33 @@ class BaseShell(Cmd):
         self._save_result(bool(args))
         return ''
 
+    def do_not(self, args: str) -> str:
+        if args:
+            return FALSE
+        return TRUE
+
     ############################################################################
     # Overrides
     ############################################################################
 
-    def onecmd(self, line: str) -> bool:
+    def onecmd(self, line: str, print_result=True) -> bool:
         """Parse and run `line`.
         Returns 0 on success and None otherwise
         """
+
         try:
             line = self.onecmd_prehook(line)
-            lines = list(self.parse_commands(line))
-            self.run_commands(lines)
+
+            if DEFINE_FUNCTION in self.locals and self.locals[DEFINE_FUNCTION].multiline:
+                self._define_multiline_function(line)
+            else:
+                lines = list(parse_commands(line,
+                                            self.delimiters,
+                                            self.ignore_invalid_syntax))
+                result = self.run_commands(lines)
+
+                if print_result and result is not None:
+                    print(result)
 
         except CancelledError:
             pass
@@ -553,9 +437,17 @@ class BaseShell(Cmd):
 
     def default(self, line: str):
         head, *tail = line.split(' ')
-        if head in self.locals['functions']:
-            f = self.locals['functions'][head]
-            return self.call_inline_function(f, *tail)
+        if head in self.env and isinstance(self.env[head], InlineFunction):
+            f = self.env[head]
+            try:
+                result = self.call_inline_function(f, *tail)
+            except ShellError:
+                # reset local scope
+                self.reset_locals()
+                raise
+
+            self.reset_locals()
+            return result
 
         if line in self._chars_allowed_for_char_method:
             return self._do_char_method(line)
@@ -576,24 +468,17 @@ class BaseShell(Cmd):
         if not lines:
             return
 
+        self.locals.set(IF, [])
+
         for i, line in enumerate(lines):
 
-            if LEFT_ASSIGNMENT not in self.locals \
-                    and i+1 < len(lines) \
-                    and '<-' in lines[i+1]:
-                # prefix line if '<-' is used later on
-                j = 1 if ';' in line else 0
-                line.insert(j, 'assign')
-
             if DEFINE_FUNCTION in self.locals:
-                if ':' in line:
-                    line = line[1:]
-
-                # self.locals[DEFINE_FUNCTION].command += '; ' + ' '.join(line)
-                self.locals[DEFINE_FUNCTION].command += ' ' + ' '.join(line)
+                self._extend_inline_function_definition(line)
                 result = None
                 continue
 
+            self._conditionally_insert_assign_operator(lines, i, line)
+
             try:
                 result = self.run_single_command(line, result)
 
@@ -613,42 +498,101 @@ class BaseShell(Cmd):
 
                 raise ShellError(str(e))
 
-            if DEFINE_FUNCTION not in self.locals:
-                # handle inline `<-`
-                if LEFT_ASSIGNMENT in self.locals and 'assign' not in line:
-                    self._save_assignee(result)
-                    result = ''
+            # handle inline `<-`
+            if DEFINE_FUNCTION not in self.locals \
+                    and LEFT_ASSIGNMENT in self.locals \
+                    and 'assign' not in line:
+                self._save_assignee(result)
+                result = ''
 
         if DEFINE_FUNCTION in self.locals:
-            self._save_inline_function()
+            if not self.locals[DEFINE_FUNCTION].multiline:
+                if self.locals[DEFINE_FUNCTION].command:
+                    self._save_inline_function()
+                else:
+                    self.locals[DEFINE_FUNCTION].multiline = True
+            return
 
         elif LEFT_ASSIGNMENT in self.locals and not io_util.interactive:
             # cancel assignment
             self._save_assignee(result)
+            return
 
-        elif result is not None:
-            print(result)
+        return result
 
-    def run_single_command(self, command_and_args: List[str], result: str = '') -> str:
-        result = self.filter_result(command_and_args, result)
+    def _conditionally_insert_assign_operator(self, lines, i, line):
+        if LEFT_ASSIGNMENT not in self.locals \
+                and i+1 < len(lines) \
+                and '<-' in lines[i+1]:
+            # prefix line if '<-' is used later on
+            j = 1 if ';' in line else 0
+            line.insert(j, 'assign')
+
+    def _extend_inline_function_definition(self, line):
+        if not self.locals[DEFINE_FUNCTION].multiline:
+            if ':' in line:
+                line = line[1:]
+
+        cmd = ' ' + ' '.join(line)
+        if self.locals[DEFINE_FUNCTION].multiline:
+            self.locals[DEFINE_FUNCTION].inner[-1] += cmd
+        else:
+            self.locals[DEFINE_FUNCTION].command += cmd
+        return line
+
+    def run_single_command(self, command_and_args: List[str], prev_result: str = '') -> str:
+        prev_result = self.filter_result(command_and_args, prev_result)
 
         prefixes, line, infix_operator_args = self.parse_single_command(
             command_and_args)
 
         if prefixes:
+            if THEN in prefixes:
+                if not self.locals[IF]:
+                    if self.ignore_invalid_syntax:
+                        return ''
+                    raise ShellError(
+                        f'If-then clause requires an {IF} statement')
+
+                self.locals[IF][-1]['depth'] += 1
+
+                if len(prefixes) <= 1 and self.locals[IF][-1]['depth'] >= 2:
+                    raise ShellError(
+                        f'If-then clause requires an {IF} statement')
+
+                if not self.locals[IF][-1]['value']:
+                    # skip
+                    return ''
+                # otherwise continue
+
             if prefixes[-1] in delimiters.bash:
-                return self.pipe_cmd_sh(line, result, delimiter=prefixes[-1])
+                return self.pipe_cmd_sh(line, prev_result, delimiter=prefixes[-1])
 
             elif prefixes[-1] == '>>=':
                 # monadic bind
                 # https://en.wikipedia.org/wiki/Monad_(functional_programming)
                 line = f'map {line}'
-                return self.pipe_cmd_py(line, result)
+                return self.pipe_cmd_py(line, prev_result)
 
             elif prefixes[-1] == RIGHT_ASSIGNMENT:
                 # TODO verify syntax of `line`
                 assert ' ' not in line
-                self.set_env_variables(line, result)
+                self.set_env_variables(line, prev_result)
+                return ''
+
+            elif prefixes[-1] == IF:
+                assert_if_then_is_closed = False
+                if assert_if_then_is_closed:
+                    if self.locals[IF] and self.locals[IF][-1]['depth'] == 0:
+                        raise ShellError(
+                            f'The previous if-then statement was not closed')
+
+                if self.locals[IF] and not self.locals[IF][-1]['value']:
+                    value = False
+                else:
+                    value = line != ''
+
+                self.locals[IF].append({'value': value, 'depth': 0})
                 return ''
 
         if infix_operator_args:
@@ -656,7 +600,7 @@ class BaseShell(Cmd):
         elif is_function_definition(command_and_args):
             return self.handle_define_inline_function(command_and_args)
 
-        return self.pipe_cmd_py(line, result)
+        return self.pipe_cmd_py(line, prev_result)
 
     def _save_assignee(self, result: str):
         keys = self.locals[LEFT_ASSIGNMENT]
@@ -697,9 +641,10 @@ class BaseShell(Cmd):
     def parse_single_command(self, command_and_args: List[str]) -> Tuple[List[str], str, List[str]]:
         # strip right-hand side delimiters
         all_args = list(omit_prefixes(command_and_args, self.delimiters))
-        f, *args = all_args
-        args = list(self.expand_variables(args))
-        line = ' '.join(chain.from_iterable(([f], args)))
+        all_args = list(expand_variables(all_args, self.env,
+                        self.completenames_options, self.ignore_invalid_syntax))
+        _f, *args = all_args
+        line = ' '.join(all_args)
 
         there_is_an_infix_operator = for_any(
             self.infix_operators, contains, args)
@@ -739,57 +684,6 @@ class BaseShell(Cmd):
         log(stderr)
         return stdout
 
-    ############################################################################
-    # Argument Parsing
-    ############################################################################
-
-    def parse_commands(self, line: str) -> Iterable[List[str]]:
-        """Split up `line` into an iterable of single commands.
-        """
-        try:
-            # split lines and handle quotes
-            # e.g. convert 'echo "echo 1"' to ['echo', 'echo 1']
-            terms = shlex.split(line, comments=True)
-
-            # re-quote delimiters
-            for i, term in enumerate(terms):
-                # if other_delimiters:
-                if term in self.delimiters or term == '=':
-                    if '"' + term + '"' in line or "'" + term + "'" in line:
-                        terms[i] = f'"{terms[i]}"'
-
-            # split `:`
-            for i, term in enumerate(terms):
-                if term.endswith(':'):
-                    # verify that the term wasn't quoted
-                    if '"' + term + '"' in line or "'" + term + "'" in line:
-                        continue
-
-                    terms[i] = terms[i][:-1]
-                    terms.insert(i+1, ':')
-                    break
-
-        except ValueError as e:
-            msg = f'Invalid syntax: {e} for {str(line)[:10]} ..'
-            if self.ignore_invalid_syntax:
-                log(msg)
-                return []
-
-            raise ShellError(msg)
-
-        if not terms:
-            return []
-
-        ################################################################################
-        # handle lines that end with `;`
-        # e.g. 'echo 1; echo 2;'
-        # TODO this doesn't preserve ; when it was originally enclosed in quotes
-        # terms = chain.from_iterable([split_tips(term.strip(), ';') for term in terms])
-        ################################################################################
-
-        # group terms based on delimiters
-        return split_sequence(terms, self.delimiters, return_delimiters=True)
-
     def infix_command(self, *args: str):
         """Treat `args` as an infix command.
         Apply the respective infix method to args.
@@ -802,7 +696,7 @@ class BaseShell(Cmd):
                 continue
 
             try:
-                lhs, rhs = self.infer_infix_args(op, *args)
+                lhs, rhs = infer_infix_args(op, *args)
             except ValueError:
                 msg = f'Invalid syntax for infix operator {op}'
                 if self.ignore_invalid_syntax:
@@ -814,89 +708,206 @@ class BaseShell(Cmd):
 
         raise ValueError()
 
-    def infer_infix_args(self, op: str, *args: str) -> Tuple[Tuple[str], Tuple[str]]:
-        if args[1] == op:
-            lhs, _, *rhs = args
-            lhs = (lhs,)
+    ############################################################################
+    # Environment Variables
+    ############################################################################
+
+    def set_env_variable(self, k: str, *values: str):
+        """Set the variable `k` to `values`
+        """
+        if not is_valid_method_name(k):
+            raise ShellError(f'Invalid variable name format: {k}')
+
+        log(f'set {k}')
+        self.env[k] = ' '.join(values)
+        return k
+
+    def set_env_variables(self, keys: Union[str, List[str]], result: str):
+        """Set the variables `keys` to the values in result.
+        """
+        if result is None:
+            raise ShellError(f'Missing return value in assignment: {keys}')
+
+        # TODO set multiple keys based on pattern matching
+        # e.g. i j = range(2)
+        if isinstance(keys, str):
+            keys = keys.split(' ')
+
+        try:
+            if len(result) == len(keys):
+                self.env.update(items=zip(keys, result))
+                return
+        except TypeError:
+            pass
+
+        if len(keys) == 1:
+            self.env[keys[0]] = result
+        elif isinstance(result, str):
+            lines = result.split('\n')
+            terms = result.split(' ')
+            if len(lines) == len(keys):
+                self.env.update(items=zip(keys, lines))
+
+            elif len(terms) == len(keys):
+                self.env.update(items=zip(keys, terms))
+
+            elif result == '':
+                self.env.update(items=zip(keys, repeat('')))
+
         else:
-            i = args.index(op)
-            lhs = args[:i]
-            rhs = args[i:]
-        return lhs, rhs
+            raise ShellError(
+                f'Cannot assign values to all keys: {" ".join(keys)}')
 
-    def expand_variables(self, terms: List[str]) -> Iterable[str]:
-        """Replace variables with their values. 
-        E.g.
-        ```sh
-        a = 1
-        print $a # gets converted to `print 1`
-        ```
-        """
-        for v in terms:
-            matches = match_words(v, prefix=r'\$')
-            if matches:
-                for match in matches:
-                    k = match[1:]
-                    if not is_valid_method_name(k):
-                        # ignore this variable silently
-                        continue
+    def handle_set_env_variable(self, lhs: Tuple[str], *rhs: str) -> str:
+        if len(lhs) != 1:
+            raise ShellError()
 
-                    error_msg = f'Variable `{match}` is not set'
-
-                    if k in self.env:
-                        v = v.replace(match, str(self.env[k]))
-                    elif self.ignore_invalid_syntax:
-                        log(error_msg)
-                    else:
-                        raise ShellError(error_msg)
-
-            if is_globbable(v):
-                try:
-                    matches = glob(v, self.completenames_options, strict=True)
-                    yield ' '.join(matches)
-                    continue
-
-                except ValueError as e:
-                    if self.ignore_invalid_syntax:
-                        log(f'Invalid syntax: {e}')
-                    else:
-                        raise ShellError(e)
-
-            yield v
-
-    def expand_variables_inline(self, line: str):
-        """Expand $variables in `line`.
-        """
-        terms = line.split(' ')
-        expanded_terms = self.expand_variables(terms)
-        line = ' '.join(expanded_terms)
-        return line
-
-    def expand_inner_variables(self, line: str):
-        matches = match_words(line, prefix=r'\$')
-        for match in matches:
-            line = line.replace(match, '')
-
-    def none(self, _: str) -> str:
-        """Do nothing. Similar to util.none.
-        This is a default value for self._do_char_method.
-        """
+        k = lhs[0]
+        self.set_env_variable(k, *rhs)
         return ''
 
-    def _save_result(self, value, overwrite=True):
-        if overwrite:
-            self._last_results = [value]
+    ############################################################################
+    # Persistency: Save/load sessions to disk
+    ############################################################################
+
+    def save_session(self, session=default_session_filename):
+        self.save_session_prehook()
+
+        if not self.env:
+            logging.info('No env data to save')
+            return
+
+        self.reset_locals()
+        env = self.locals[ENV]
+
+        with open(session, 'w') as f:
+            try:
+                json = dumps(env)
+            except TypeError:
+                logging.debug('Cannot serialize self.env')
+                try:
+                    json = dumps(env, skip_keys=True)
+                except TypeError:
+                    json = dumps(asdict(env))
+
+            f.write(json)
+
+    def reset_locals(self):
+        self.locals.cd()
+
+    def try_load_session(self, session=default_session_filename):
+        self.load_session(session, strict=False)
+
+    def load_session(self, session: str = None, strict=True):
+        try:
+            with open(session) as f:
+                data = f.read()
+
+        except OSError as e:
+            if strict:
+                raise ShellError(e)
+
+            log(f'Session file not found: {session}: {e}')
+            return
+
+        if not data:
+            logging.info('No env data found')
+            return
+
+        env = loads(data)
+
+        self.reset_locals()
+        log(f'Using session: {session}')
+        show(env)
+
+        # TODO handle key conflicts
+        self.env.update(env)
+
+        self.load_session_posthook()
+
+    ############################################################################
+    # Inline & Multiline Functions
+    ############################################################################
+
+    def call_inline_function(self, f: InlineFunction, *args: str):
+        translations = {}
+
+        if len(args) != len(f.args):
+            raise ShellError(
+                f'Invalid number of arguments: {len(f.args)} arguments expected .')
+
+        for i, k in enumerate(f.args):
+            # quote item to preserve `\n`
+            translations[k] = shlex.quote(args[i])
+
+        with enter_new_scope(self):
+
+            for line in f.inner:
+                terms = [term for term in line.split(' ') if term != '']
+                terms = list(translate_terms(terms, translations))
+
+                self.onecmd(' '.join(terms), print_result=False)
+
+            terms = [term for term in f.command.split(' ') if term != '']
+            terms = list(translate_terms(terms, translations))
+
+            first_func = terms[0]
+            if not self.is_function(first_func):
+                terms = ['print'] + terms
+
+            result = self.eval(terms, quote=False)
+
+        return result
+
+    def handle_define_inline_function(self, terms: List[str]) -> str:
+        f, *args = terms
+        args = [arg for arg in args if arg != '']
+
+        if not args or not args[0].startswith('(') or not args[-1].endswith(')'):
+            lhs = ' '.join(terms)
+            raise ShellError(f'Invalid syntax for inline function: {lhs}')
+
+        # strip braces
+        if args[0] == '()':
+            args = []
         else:
-            self._last_results.append(value)
+            args[0] = args[0][1:]
+            args[-1] = args[-1][:-1]
 
-    def translate_terms(self, terms: Iterable[str], translations: dict):
-        for term in terms:
-            term = term.strip()
-            if term in translations:
-                yield str(translations[term])
-                continue
+        self.locals.set(DEFINE_FUNCTION,
+                        InlineFunction('', *args, func_name=f))
 
-            yield term
+    def _define_multiline_function(self, line: str):
+        line = line.lstrip()
+        if line.startswith(RETURN + ' '):
+            line = removeprefix(line, RETURN + ' ')
+            self.locals[DEFINE_FUNCTION].command = line
+            self._save_inline_function()
+
+        else:
+            self.locals[DEFINE_FUNCTION].inner.append(line)
+
+    def _save_inline_function(self) -> str:
+        func = self.locals[DEFINE_FUNCTION]
+        self.locals.rm(DEFINE_FUNCTION)
+
+        f = func.func_name
+        if has_method(self, f'do_{f}'):
+            raise ShellError(
+                f'Name conflict: Cannot define inline function {f}, '
+                f'because there already exists a method do_{f}.')
+
+        if not is_valid_method_name(f):
+            raise ShellError(f'Invalid function name format: {f}')
+
+        if not func.multiline:
+            func.command = expand_variables_inline(func.command, self.env,
+                                                   self.completenames_options,
+                                                   self.ignore_invalid_syntax)
+
+        self.env[f] = func
+        positionals = ' '.join(func.args)
+        log(f'function {f}({positionals});')
 
 
 def is_function_definition(terms: List[str]) -> bool:
@@ -913,3 +924,21 @@ def is_function_definition(terms: List[str]) -> bool:
         _f, first, *_, last = terms
 
     return first.startswith('(') and last.endswith(')')
+
+
+def scope() -> dict:
+    return defaultdict(dict)
+
+
+@contextmanager
+def enter_new_scope(cls: BaseShell, scope_name=INNER_SCOPE):
+    """Create a new scope, then change directory into that scope. 
+    Finally exit the new scope.
+    """
+    cls.locals.set(scope_name, scope())
+    try:
+        with cd(cls.locals, scope_name):
+            cls.init_current_scope()
+            yield
+    finally:
+        pass
