@@ -2,6 +2,9 @@ from enum import Enum
 
 
 class Python(Enum):
+    RETURN = 'return'
+    IF = 'if'
+    THEN = 'then'
     NEW_COMMAND = ';'
     DEFINE_FUNCTION = ':'
     RIGHT_ASSIGNMENT = '->'
@@ -9,6 +12,9 @@ class Python(Enum):
     PIPE = '|>'
     MAP = '>>='
     SET_ENV_VARIABLE = '='
+    # ELSE = 'else'
+    # AND = 'and'
+    # OR = 'or'
 
 
 bash = ['|', '>', '>>', '1>', '1>>', '2>', '2>>']
@@ -16,6 +22,17 @@ python = [o.value for o in Python]
 all = bash + python
 
 # use dedicated variables to simplify searching
+RETURN = Python.RETURN.value
 DEFINE_FUNCTION = Python.DEFINE_FUNCTION.value
 RIGHT_ASSIGNMENT = Python.RIGHT_ASSIGNMENT.value
 LEFT_ASSIGNMENT = Python.LEFT_ASSIGNMENT.value
+IF = Python.IF.value
+THEN = Python.THEN.value
+# ELSE = Python.ELSE.value
+# AND = Python.AND.value
+# OR = Python.OR.value
+
+
+class KeyWords(Enum):
+    AND = 'and'
+    OR = 'or'
