@@ -285,6 +285,18 @@ def split_prefixes(items: Sequence[T], prefixes: Sequence[T]) -> Iterable[T]:
     return takewhile(predicate, items)
 
 
+def translate_terms(terms: Iterable[str], translations: dict):
+    """Iterate over `terms` and return any translations found.
+    """
+    for term in terms:
+        term = term.strip()
+        if term in translations:
+            yield str(translations[term])
+            continue
+
+        yield term
+
+
 def omit_prefixes(items: Sequence[T], prefixes: Sequence[T]) -> Iterable[T]:
     predicate = partial(contains, prefixes)
     return dropwhile(predicate, items)
@@ -533,6 +545,13 @@ def call(f, *_):
     """
     return f()
 
+
+def removeprefix(s: str, prefix: str) -> str:
+    # from python 3.11
+    if s.startswith(prefix):
+        return s[len(prefix):]
+    return s
+
 ################################################################################
 # Predicates
 ################################################################################
